import typing

from fastapi import Request, HTTPException
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse

from ytservice.common import build_response, RESULT_CODE_ERROR


# Helper function to format exception messages
def format_exception_message(exc: typing.Any) -> str:
    if isinstance(exc, RequestValidationError):
        return "; ".join([error['msg'] for error in exc.errors()])
    elif isinstance(exc, HTTPException):
        return exc.detail
    return "Internal server error"


# Exception handler for validation errors
async def validation_exception_handler(request: Request, exc: RequestValidationError):
    return JSONResponse(
        status_code=422,
        content=build_response(code=RESULT_CODE_ERROR, message=format_exception_message(exc))
    )


# Exception handler for HTTP exceptions
async def http_exception_handler(request: Request, exc: HTTPException):
    return JSONResponse(
        status_code=exc.status_code,
        content=build_response(code=RESULT_CODE_ERROR, message=format_exception_message(exc))
    )


# Exception handler for general exceptions
async def global_exception_handler(request: Request, exc: Exception):
    return JSONResponse(
        status_code=500,
        content=build_response(code=RESULT_CODE_ERROR, message=format_exception_message(exc))
    )



from pathlib import Path

from fastapi import HTTPException
from pydantic import BaseModel, field_validator
from pydantic_settings import BaseSettings

SETTING_HOST = "0.0.0.0"
SETTING_PORT = 8000
SETTING_PROXY = "socks5://127.0.0.1:1086"


# Configuration settings for the application
class Settings(BaseSettings):
    host: str = SETTING_HOST
    port: int = SETTING_PORT
    proxy: str = SETTING_PROXY
    download_dir: Path = Path.home() / "YTDownload"
    home_dir: Path = Path.home() / ".yt-home"


# Model for YouTube URL input
class YtModel(BaseModel):
    url: str

    # Validator to ensure the URL starts with http:// or https://
    @field_validator('url')
    def url_valid(cls, v):
        if not (v.startswith("http://") or v.startswith("https://")):
            raise HTTPException(status_code=500, detail='URL must start with http:// or https://')
        return v



import sys
import urllib.parse
from pathlib import Path

import click
import uvicorn
from fastapi import FastAPI, HTTPException
from fastapi.exceptions import RequestValidationError
from fastapi.responses import FileResponse

from ytservice import (model, ytclient, common, exception_handlers)
from ytservice.model import (Settings, YtModel)
from ytservice.version import VERSION

APP_NAME = "yt-service"

# Initialize settings and FastAPI app
settings = Settings()
app = FastAPI(redoc_url=None)

app.add_exception_handler(RequestValidationError, exception_handlers.validation_exception_handler)
app.add_exception_handler(HTTPException, exception_handlers.http_exception_handler)
app.add_exception_handler(Exception, exception_handlers.global_exception_handler)


@app.get("/")
def read_root():
    return dict(app=APP_NAME, version=VERSION,
                Lan="http://{0}:{1}".format(common.get_local_ip(), settings.port))


# Endpoint to extract audio information
@app.post("/yt/audio/info")
async def audio_info(post_model: YtModel):
    try:
        content = ytclient.extract_audio_info(url=post_model.url, settings=settings)
        if content:
            return common.build_response(data=content)
        else:
            raise HTTPException(status_code=500, detail="Failed to extract video info")
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# Endpoint to extract audio information
@app.post("/yt/audio/bestInfo")
async def audio_best_info(post_model: YtModel):
    try:
        content = ytclient.optimize_audio_info(url=post_model.url, settings=settings)
        if content:
            return common.build_response(data=content)
        else:
            raise HTTPException(status_code=500, detail="Failed to extract video bestInfo")
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# Endpoint to download video
@app.post("/yt/video/download")
async def video_download(post_model: YtModel):
    try:
        download_info = ytclient.download_video(url=post_model.url, settings=settings)
        if download_info:
            download_url = download_info["url"]
            file_name = Path(download_url).name
            encoded_file_name = urllib.parse.quote(file_name)
            # Include timestamp directory in the download URL
            timestamp = Path(download_url).parent.name
            file_url = f"http://{common.get_local_ip()}:{settings.port}/yt/download-file/{timestamp}/{encoded_file_name}"
            download_info["url"] = file_url
            return common.build_response(data=download_info)
        else:
            raise HTTPException(status_code=500, detail="Failed to download video")
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# Endpoint to extract video information
@app.post("/yt/video/info")
async def video_info(post_model: YtModel):
    try:
        content = ytclient.extract_video_info(url=post_model.url, settings=settings)
        if content:
            return common.build_response(data=content)
        else:
            raise HTTPException(status_code=500, detail="Failed to extract video info")
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# Endpoint to extract video information
@app.post("/yt/video/bestInfo")
async def video_best_info(post_model: YtModel):
    try:
        content = ytclient.optimize_video_info(url=post_model.url, settings=settings)
        if content:
            return common.build_response(data=content)
        else:
            raise HTTPException(status_code=500, detail="Failed to extract video bestInfo")
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# Endpoint to serve downloaded files
@app.get("/yt/download-file/{timestamp}/{encoded_file_name}")
async def download_file(timestamp: str, encoded_file_name: str):
    file_name = urllib.parse.unquote(encoded_file_name)
    file_location = settings.download_dir / timestamp / file_name
    if not file_location.exists():
        raise HTTPException(status_code=404, detail="File not found")
    return FileResponse(path=file_location, filename=file_location.name)


# CLI command to start the server with optional settings
@click.command()
@click.option('--host', type=str, show_default=True, default=model.SETTING_HOST, help='Service IP address')
@click.option('--port', type=int, show_default=True, default=model.SETTING_PORT, help='Service port')
@click.option('--proxy', type=str, show_default=True, default=model.SETTING_PROXY, help='Proxy server address')
@click.option('--download-dir', type=click.Path(), help='Download directory [~/YTDownload]')
@click.option('-s', '--signal', help='send signal to service',
              type=click.Choice(['install', 'start', 'stop', 'status', 'rm'], case_sensitive=False))
@click.option('--version', is_flag=True)
def start_server(host, port, proxy, download_dir, signal, version):
    if version:
        click.echo(VERSION)
        return
    if signal:
        common.run_service(app_name=APP_NAME, signal=signal)
        return
    if host:
        settings.host = host
    if port:
        settings.port = port
    if proxy:
        settings.proxy = proxy
    if download_dir:
        settings.download_dir = Path(download_dir)
    uvicorn.run(app, host=settings.host, port=settings.port)


# Main entry point to start the server
if __name__ == "__main__":
    sys.argv.append("--version")
    start_server()
